import functools


def on_stage(stage: str):
    """decorator for callback methods so that they only run on a certain stage"""

    def wrapper(f):
        @functools.wraps(f)
        def wrapped(self, loop, *args, **kwargs):
            if loop.stage == stage:
                return f(self, loop, *args, **kwargs)

        return wrapped

    return wrapper


def if_active(f):
    """decorator for callback methods so that they are only called when active"""

    @functools.wraps(f)
    def inner(self, loop, *args, **kwargs):
        if self.active:
            return f(self, loop, *args, **kwargs)

    return inner


def if_inactive(f):
    """decorator for callback methods so that they are only called when inactive"""

    @functools.wraps(f)
    def inner(self, loop, *args, **kwargs):
        if not self.active:
            return f(self, loop, *args, **kwargs)

    return inner



from typing import Optional, Union, Dict, Mapping, Callable
import torch
from torch import nn
from hearth.containers import TensorDict, NumberDict
from hearth._multihead import _MultiHeadFunc


def _identity(x):
    return x


class MultiHeadLoss(nn.Module, _MultiHeadFunc):
    """wrapper for losses for models with multiple output heads.

    Args:
        weights: a mapping of key to scalar weight, will be multiplied with losses before summing
            to create the aggregate value at `aggregate_key`, need not sum to 1. Defaults to None
            (all losses weighted evenly).
        aggregate_key: the key to use for the aggregate loss. Defaults to 'weighted_sum'.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.losses import MultiHeadLoss
        >>>
        >>> _ =torch.manual_seed(0)
        >>>
        >>> loss = MultiHeadLoss(a=nn.BCEWithLogitsLoss(),
        ...                      b=nn.CrossEntropyLoss())
        >>> loss
        MultiHeadLoss(a=BCEWithLogitsLoss(),
                      b=CrossEntropyLoss(),
                      weights=NumberDict({'a': 1.0, 'b': 1.0}),
                      aggregate_key=weighted_sum)

        multihead loss expects inputs and targets to be dicts with containing
        the it's keys (in this case 'a' and 'b'):

        >>> batch_size = 10
        >>> inputs = {'a': torch.rand(batch_size, 1),
        ...           'b': torch.normal(batch_size, 1, size=(batch_size, 4))}
        >>> targets = {'a': torch.rand(batch_size, 1).round(),
        ...            'b': torch.randint(4, size=(batch_size,))}
        >>>
        >>> loss(inputs, targets)
        TensorDict({'a': tensor(0.5791), 'b': tensor(1.2425), 'weighted_sum': tensor(1.8216)})

        ``weighted_sum`` is the default aggregate key. this is the bit you should call backward on.
        you can change the default aggregate key if you like by specifying it at init.

        >>> loss = MultiHeadLoss(a=nn.BCEWithLogitsLoss(),
        ...                      b=nn.CrossEntropyLoss(),
        ...                      aggregate_key='sally')
        >>> loss(inputs, targets)
        TensorDict({'a': tensor(0.5791), 'b': tensor(1.2425), 'sally': tensor(1.8216)})

        you can aslo specify ``weights`` at init to weight contribution losses differently:

        >>> loss = MultiHeadLoss(a=nn.BCEWithLogitsLoss(),
        ...                      b=nn.CrossEntropyLoss(),
        ...                      weights={'a': .2, 'b':.8})
        >>> loss(inputs, targets)
        TensorDict({'a': tensor(0.5791), 'b': tensor(1.2425), 'weighted_sum': tensor(1.1098)})
    """

    def __init__(
        self,
        *,
        weights: Optional[Union[NumberDict, Dict[str, int]]] = None,
        aggregate_key: str = 'weighted_sum',
        **kwargs,
    ):
        nn.Module.__init__(self)
        self._fns = nn.ModuleDict(kwargs)
        self.aggregate_key = aggregate_key
        self.weights = weights

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, weights: Optional[Union[NumberDict, Dict[str, int]]] = None):
        if weights is not None:
            if weights is not NumberDict:
                weights = NumberDict(weights)
            if set(weights.keys()) != set(self.keys()):
                raise ValueError('weight keys must match keys for loss functions!')
        else:
            weights = NumberDict({k: 1.0 for k in self.keys()})
        self._weights = weights

    def _aggregate(self, out):
        out[self.aggregate_key] = (out * self.weights).sum()
        return out

    def forward(
        self, inputs: Mapping[str, torch.Tensor], targets: Mapping[str, torch.Tensor], **kwargs
    ) -> TensorDict:
        out = TensorDict()
        for k, func in self.items():
            out[k] = func(inputs[k], targets[k], **kwargs)
        return self._aggregate(out)

    def _argrepr(self):
        return (
            super()._argrepr() + f', weights={self.weights!r}, aggregate_key={self.aggregate_key}'
        )

    def __repr__(self):
        return f'{self.__class__.__name__}({self._argrepr()})'


class _BaseLoss(nn.Module):
    """base class for losses."""

    _reductions: Dict[str, Callable] = {'mean': torch.mean, 'sum': torch.sum, 'none': _identity}

    def __init__(self, *args, reduction: str, **kwargs):
        super().__init__()
        if reduction not in self._reductions:
            raise ValueError(
                f'reduction {reduction!r} is not supported for {self.__class__.__name__},'
                f' please choose one of {list(self._reductions)!r}'
            )
        self.reduction = reduction
        self._reduce_fn = self._reductions[reduction]

    def _reduce(self, x: torch.Tensor) -> torch.Tensor:
        return self._reduce_fn(x)

    def extra_repr(self) -> str:
        return f'reduction={self.reduction}'


class _MaskedLoss(_BaseLoss):
    """base class for losses that support masking by target value."""

    def __init__(self, *args, mask_target_value: int = -1, reduction: str, **kwargs):
        super().__init__(*args, reduction=reduction, **kwargs)
        self.mask_target_value = mask_target_value

    def _masked_reduce(self, x, mask: torch.Tensor) -> torch.Tensor:
        if self.reduction == 'none':
            return x * (mask * 1.0)
        return self._reduce_fn(x[mask])

    def _get_mask(self, targets: torch.Tensor) -> torch.Tensor:
        return targets != self.mask_target_value

    def extra_repr(self) -> str:
        return f'mask_target_value={self.mask_target_value}, reduction={self.reduction!r}'


class MulticlassFocalLoss(_MaskedLoss):
    """multiclass focal loss.

    Focal loss is a proposed solution for handling class imbalambce in segmentation.

    Reference:
       `Li et al. : Focal Loss for Dense Object Detection <https://arxiv.org/abs/1708.02002>`_

    Args:
        alpha: class weighting factor, may be a scalar, or a tensor with one weight per class
            if None alpha=1.0. Defaults to None.
        gamma: focusing factor. Defaults to 2.0.
        mask_target_value: this index will be masked when seen in the targets upon
            computing the loss. Defaults to -1.
        reduction: string . Defaults to 'mean'.

    Shape:
        - inputs: :math:`(N, *)` where :math:`*` means, any number of additional dimensions
        - targets: :math:`(N, *, C)`, where `C = number of classes`.
        - output: scalar unless :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same shape as\
            targets.

    Example:
        >>> import torch
        >>> from hearth.losses import MulticlassFocalLoss
        >>>
        >>> inp = torch.tensor([[ 1.3053, -0.6421, -1.2027,  1.0494, -1.9540],
        ...                     [ 0.6801,  0.2266,  0.8120,  0.9490, -0.8120],
        ...                     [-0.8212,  0.8024,  0.8370, -0.3272,  0.6125],
        ...                     [-0.1975, -1.0706,  2.6819, -0.4297,  0.1980],
        ...                     [ 0.8256,  1.7839, -1.5876,  1.7705, -1.7051],
        ...                     [ 0.1288,  1.0981,  0.0570, -1.1684,  0.4567],
        ...                     [ 0.5658,  1.3948, -1.1457, -0.5921, -0.8026],
        ...                     [-0.3989,  0.6574,  0.3411, -1.9814,  0.2935]])
        >>>
        >>> targets = torch.tensor([0, 1, 4, 2, 3, 0, 2, 2])
        >>> loss = MulticlassFocalLoss()
        >>> loss
        MulticlassFocalLoss(alpha=1.0, gamma=2.0, mask_target_value=-1, reduction='mean')

        >>> loss(inp, targets)
        tensor(0.9478)

        pass class weights for the alpha value:

        >>> weights = torch.tensor([1.0, 2.0, 0.3, 2.1, .5])
        >>> loss = MulticlassFocalLoss(alpha=weights)
        >>> loss(inp, targets)
        tensor(0.8010)

        supports masking by value for instance if we want to predict classes over time and the last\
         two timesteps for the last batch are missing we can mask them by setting to \
         :attr:`mask_target_value` to -1 and padding targets with -1:

        >>> masked_targets =  torch.tensor([[0, 1, 4, 2], [3, 0, -1, -1]]) # (batch, time)
        >>> inp = inp.reshape((2, 4, -1)) # (batch, time, classes)
        >>> loss(inp, masked_targets)
        tensor(0.8885)

        when :attr:`reduction` is ``'none'`` we will return one loss value per step as you can see\
         the masked targets are 0.0:

        >>> loss = MulticlassFocalLoss(alpha=weights, reduction='none')
        >>> loss(inp, masked_targets)
        tensor([[1.8430e-01, 2.7830e+00, 4.0199e-01, 1.6719e-03],
                [6.7119e-01, 1.2889e+00, 0.0, 0.0]])
    """

    def __init__(
        self,
        alpha: Optional[Union[torch.Tensor, float]] = None,
        gamma: float = 2.0,
        mask_target_value: int = -1,
        reduction: str = 'mean',
    ):
        super().__init__(mask_target_value=mask_target_value, reduction=reduction)
        self.gamma = gamma
        self.alpha = alpha

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha: Union[torch.Tensor, float, None]):
        if alpha is None:
            alpha = 1.0
        if isinstance(alpha, float):
            self._alpha = alpha
            self._scalar_alpha = True
        elif isinstance(alpha, torch.Tensor):
            self.register_buffer('_alpha', alpha)
            self._scalar_alpha = False

    def _get_alphas(self, targets: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        if self._scalar_alpha:
            return self.alpha

        alpha_t = self.alpha[targets]
        return alpha_t * (mask * 1.0)

    def _get_ce(self, inputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        if len(inputs.shape) > 2:
            # this torch expects (batch, classes, ...)
            # where we expect (batch, ..., classes)
            # so we transpose before throwing in
            inputs = inputs.transpose(1, -1)

        return nn.functional.cross_entropy(
            inputs, targets, reduction='none', ignore_index=self.mask_target_value
        )

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        ce = self._get_ce(inputs, targets)
        mask = self._get_mask(targets)
        p_t = torch.exp(-ce)
        alpha_t = self._get_alphas(targets, mask)
        focal_loss = alpha_t * (1 - p_t) ** self.gamma * ce
        return self._masked_reduce(focal_loss, mask=mask)

    def extra_repr(self) -> str:
        parent_args = super().extra_repr()
        return f'alpha={self.alpha!r}, gamma={self.gamma}, {parent_args}'


class BinaryFocalLoss(_MaskedLoss):
    """Binary focal loss.

    Focal loss is a proposed solution for handling class imbalambce in segmentation.

    Reference:
       `Li et al. : Focal Loss for Dense Object Detection <https://arxiv.org/abs/1708.02002>`_

    Args:
        alpha: class weighting factor. Defaults to 0.25 (defaults taken from paper).
        gamma: focusing factor. Defaults to 2.0 (default taken from paper).
        mask_target_value: this index will be masked when seen in the targets upon
            computing the loss. Defaults to -1.
        reduction: string name of reduction. Defaults to 'mean'.

    Shape:
        - inputs: :math:`(N, *)` where :math:`*` means, any number of additional dimensions\
            if inputs have an extra single dimension thats ok... they will be reshaped as targets.
        - targets: :math:`(N, *)`.
        - output: scalar unless :attr:`reduction` is ``'none'``, then :math:`(N, *)`, same shape as\
            targets.

    Example:
        >>> import torch
        >>> from hearth.losses import BinaryFocalLoss
        >>>
        >>> targets = torch.tensor([0, 1, 0, 0, 1, 0], dtype=torch.float32)
        >>> inputs = torch.tensor([-1.1645,  -0.2928, -0.5685, -0.8038, -0.0211,  2.0062])
        >>> loss = BinaryFocalLoss()
        >>> loss
        BinaryFocalLoss(alpha=0.25, gamma=2.0, mask_target_value=-1, reduction='mean')

        >>> loss(inputs, targets)
        tensor(0.2399)

        if your inputs have an extra dim thats ok:

        >>> loss(inputs.unsqueeze(-1), targets)
        tensor(0.2399)

        supports masking by value for instance if we want to predict a binary value over time and \
        the last timestep for the last example is missing, we can mask by setting \
        :attr:`mask_target_value` to -1 and padding targets with -1:

        >>> masked_targets =  torch.tensor([[0, 1, 0], [0, 1, -1]], dtype=torch.float32)
        >>> inputs = inputs.reshape((2, 3, -1)) # (batch, time, 1)
        >>> loss(inputs, masked_targets)
        tensor(0.0393)

        when :attr:`reduction` is ``'none'`` we will return one loss value per step as you can see\
         the masked targets are 0.0:

        >>> loss = BinaryFocalLoss(reduction='none')
        >>> loss(inputs, masked_targets)
        tensor([[0.0115, 0.0697, 0.0440],
                [0.0265, 0.0449, 0.0000]])
    """

    def __init__(
        self,
        alpha: float = 0.25,
        gamma: float = 2.0,
        mask_target_value: int = -1,
        reduction: str = 'mean',
    ):
        super().__init__(mask_target_value=mask_target_value, reduction=reduction)
        self.gamma = gamma
        self.alpha = alpha

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        inputs = inputs.reshape_as(targets)
        bce = torch.nn.functional.binary_cross_entropy_with_logits(
            inputs, targets, reduction="none"
        )
        a_t = (1 - self.alpha) + targets * (2 * self.alpha - 1)
        p_t = torch.exp(-bce)
        focal = a_t * (1 - p_t) ** self.gamma * bce
        return self._masked_reduce(focal, mask=self._get_mask(targets))

    def extra_repr(self) -> str:
        parent_args = super().extra_repr()
        return f'alpha={self.alpha!r}, gamma={self.gamma}, {parent_args}'



from typing import Union, Sized
from dataclasses import dataclass
from torch.utils.data import Dataset

from hearth.containers import TensorDict


@dataclass
class XYDataset(Dataset):
    """basic dataset that returns a tuple of inputs and targets.

    supports :class:`hearth.containers.TensorDataset`
    """

    x: Union[Sized, TensorDict]
    y: Union[Sized, TensorDict]

    def __len__(self) -> int:
        if isinstance(self.x, TensorDict):
            return len(next(iter(self.x.values())))  # type: ignore
        return len(self.x)

    def __getitem__(self, index):
        return self.x[index], self.y[index]



"""transforms are basic operations that can be composed as part of a `Pipeline`_
"""
from abc import ABC, abstractmethod
from typing import TypeVar, Optional, Union, Generic

import torch
import numpy as np

InT = TypeVar('InT')
OutT = TypeVar('OutT')
TensorApplicable = Union[torch.Tensor, np.ndarray, int, float]


class Transform(ABC, Generic[InT, OutT]):
    """Abstract base class for all transforms."""

    def _repr_args(self):
        return ''

    def __repr__(self):
        return f'{self.__class__.__name__}({self._repr_args()})'

    @abstractmethod
    def __call__(self, x: InT) -> OutT:
        return NotImplemented


class Tensorize(Transform[InT, torch.Tensor]):
    """Tensorizes the given input with optional dtype and device

    Args:
        dtype : an optional string or torch.dtype. Defaults to None.
        device : [description]. Defaults to 'cpu'.

    Example:
        >>> import torch
        >>> from hearth.transforms import Tensorize
        >>>
        >>> transform = Tensorize(dtype='float32')
        >>> transform([1.1, 2.2, 3.3])
        tensor([1.1000, 2.2000, 3.3000])
    """

    def __init__(
        self,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Union[str, torch.device] = 'cpu',
    ):
        self._dtype = self._get_dtype(dtype)
        self._device = device

    def _get_dtype(self, dtype):
        if dtype is None:
            return dtype

        try:
            if isinstance(dtype, str):
                dtype = getattr(torch, dtype)
            assert isinstance(dtype, torch.dtype)
            return dtype
        except (AttributeError, AssertionError):
            raise TypeError(
                'dtype must be a valid torch.dtype or a string'
                f' corresponding to a valid torch.dtype got {dtype}.'
            )

    def _repr_args(self):
        return f'dtype={self._dtype}, device={self._device}'

    def __call__(self, x: InT) -> torch.Tensor:
        return torch.tensor(x, dtype=self._dtype, device=self._device)


class Normalize(Transform):
    """Normalize a tensor or array from a fixed mean and std

    Args:
        mean : may be float, tensor or array. if has dimensions (such as channels for images)
            must match shape of std
        std : may be float, tensor or array. if has dimensions (such as channels for images)
            must match shape of std

    Example:
        >>> import torch
        >>> from hearth.transforms import Normalize
        >>>
        >>> transform = Normalize(mean=1.5, std=1.1859)
        >>> x = torch.linspace(0, 3, 5)
        >>> transform(x)
        tensor([-1.2649, -0.6324,  0.0000,  0.6324,  1.2649])

        >>> channel_transform = Normalize(mean=torch.tensor([7.6596, 8.0000, 8.3404]),
        ...                                 std=torch.tensor([4.8622, 4.8622, 4.8622]))
        >>> x= torch.linspace(0, 16, 48).reshape(4, 4, 3)
        >>> y = channel_transform(x)
        >>> y.shape
        torch.Size([4, 4, 3])

        >>> y.mean(dim=(0, 1))
        tensor([-0.00,  0.00,  0.00])

        >>> y.std(dim=(0, 1))
        tensor([1.0000, 1.0000, 1.0000])
    """

    def __init__(self, mean: TensorApplicable, std: TensorApplicable):
        self.mean = mean
        self.std = std

    def _repr_args(self):
        return f'mean={self.mean}, std={self.std}'

    def __call__(self, x):
        return (x - self.mean) / self.std


class Pipeline(Transform):
    """Pipeline applies a chain of transforms to an input in order.

    Example:
        >>> import torch
        >>> import numpy as np
        >>> from hearth.transforms import Normalize, Tensorize, Pipeline
        >>>
        >>> pipeline = Pipeline(Tensorize(dtype='float32'), Normalize(mean=-0.34, std=1.75))
        >>> pipeline
        Pipeline(Tensorize(dtype=torch.float32, device=cpu), Normalize(mean=-0.34, std=1.75))

        >>> len(pipeline)
        2

        >>> x = np.array([-3.0, -1.5, .3, .4, 2.1])
        >>> pipeline(x)
        tensor([-1.5200, -0.6629,  0.3657,  0.4229,  1.3943])
    """

    def __init__(self, *transforms):
        self._transforms = transforms

    def _repr_args(self):
        return ', '.join(map('{!r}'.format, self._transforms))

    def __len__(self) -> int:
        return len(self._transforms)

    def __getitem__(self, i):
        return self._transforms[i]

    def __call__(self, x):
        for transform in self:
            x = transform(x)
        return x



from abc import ABC, abstractclassmethod
from copy import deepcopy
from functools import singledispatch
import inspect
from inspect import Parameter
from torch import nn
import functools
import importlib

_global_param_exclude = ('device', 'dtype')


class SupportsConfig(ABC):
    """protocoll for classes that support a set of methods for load and save from json config."""

    @classmethod
    def from_config(cls, config):
        """load an instance of the class from a config dict."""
        return from_config(cls, config)

    @abstractclassmethod
    def config(self):
        """this method should return a **jsonable** dictionary that can be used to initilize this\
         class"""
        return {}

    @classmethod
    def __subclasshook__(cls, other):
        if cls is SupportsConfig:
            try:
                return (
                    hasattr(cls, 'config')
                    and callable(cls.config)
                    and hasattr(cls, 'from_config')
                    and callable(cls.from_config)
                )
            except AttributeError:
                return False
        return NotImplemented


def import_cls(name: str, module: str, **kwargs):
    return getattr(importlib.import_module(module), name)


def _parse_args(args, signature):  # noqa: C901
    def parse_arg(arg):
        if isinstance(arg, nn.Module):
            return extended_config(arg)
        return arg

    for name, arg in args:
        sig_param = signature.parameters[name]
        if sig_param.kind == Parameter.VAR_POSITIONAL:
            yield f'*{name}', [parse_arg(a) for a in arg]

        elif sig_param.kind == Parameter.VAR_KEYWORD:
            yield from ((k, parse_arg(v)) for k, v in arg.items())

        else:
            yield name, parse_arg(arg)


def _init_wrapper(init_fn):
    def init(self, *args, **kwargs):
        sig = inspect.signature(self.__class__)
        bound = sig.bind(*args, **kwargs)
        bound.apply_defaults()
        if not hasattr(self, '__config__'):
            self.__config__ = {}
            for k, v in _parse_args(bound.arguments.items(), sig):
                if k not in self.__config__:
                    self.__config__[k] = v

        init_fn(self, *args, **kwargs)

    init_ = functools.wraps(init_fn)(init)
    return init_


def from_config(cls, config):
    def load_item(x):
        if is_extended_config(x):
            return from_extended_config(x)
        return x

    config = deepcopy(config)
    varargs = list(filter(lambda x: x.startswith('*'), config))
    if varargs:
        args = config.pop(varargs[0])
    else:
        args = []
    return cls(*map(load_item, args), **{k: load_item(v) for k, v in config.items()})


def from_extended_config(extended_config):
    if not is_extended_config(extended_config):
        raise ValueError('config does not contain nessisary metadata')
    cls = import_cls(**extended_config['meta'])

    return from_config(cls, extended_config['config'])


def get_meta(cls):
    return {'name': cls.__name__, 'module': cls.__module__}


def is_extended_config(config) -> bool:
    return (
        isinstance(config, dict)
        and ('meta' in config)
        and ('name' in config['meta'])
        and ('module' in config['meta'])
    )


def gather_params_from_instance(inst, params=None, exclude=()):
    if params is None:
        params = inspect.signature(inst.__class__).parameters
    return {
        param: getattr(inst, param)
        for param in params
        if param not in _global_param_exclude + exclude
    }


def extended_config(obj):
    meta = get_meta(obj.__class__)
    conf = config(obj)
    return {'meta': meta, 'config': conf}


@singledispatch
def config(obj):
    if hasattr(obj, 'config'):
        return obj.config()
    return NotImplemented


@config.register(nn.Module)
def module_config(mod):
    if hasattr(mod, 'config'):
        return mod.config()
    return gather_params_from_instance(mod)


@config.register(nn.Bilinear)
@config.register(nn.Conv1d)
@config.register(nn.Conv2d)
@config.register(nn.Conv3d)
@config.register(nn.ConvTranspose1d)
@config.register(nn.ConvTranspose2d)
@config.register(nn.ConvTranspose3d)
@config.register(nn.Linear)
def handle_bias(mod):
    config = gather_params_from_instance(mod, exclude=('bias',))
    config['bias'] = mod.bias is not None
    return config


@config.register(nn.Embedding)
@config.register(nn.EmbeddingBag)
def handle_embedding(mod):
    return gather_params_from_instance(mod, exclude=('_weight',))


@config.register(nn.FractionalMaxPool2d)
@config.register(nn.FractionalMaxPool3d)
def handle_frac_max_pool(mod):
    return gather_params_from_instance(mod, exclude=('_random_samples',))


@config.register(nn.GRU)
@config.register(nn.LSTM)
@config.register(nn.RNN)
def handle_rnn(mod):
    params = inspect.signature(mod.__class__.__bases__[0]).parameters
    return gather_params_from_instance(mod, params, exclude=('mode',))


@config.register(nn.Identity)
def handle_identity(mod):
    return {}


@config.register(nn.Sequential)
def handle_sequential(mod):
    return {'*args': [extended_config(m) for m in mod]}



from abc import ABC
from typing import Tuple, Mapping
import torch
from torch import Tensor
from hearth._internals import to_snakecase
from hearth.containers import TensorDict
from hearth._multihead import _MultiHeadFunc


class Metric(ABC):
    """abstract base class for all metrics.

    Note:
        Metrics should inherit from this method and define a forward method (for compatability
        with torch losses and modules)
    """

    def forward(self, inputs, target, **kwargs):
        """given call this metric given an input and target and optional keyword arguments.

        Args:
            inp: the input tensor... generally some form of prediciton.
            target: the target tensor.

        Returns:
            a scalar tensor
        """
        return NotImplemented

    def _mask(self, inputs, targets) -> Tuple[Tensor, Tensor]:
        return inputs, targets

    def _prepare(self, inputs, targets) -> Tuple[Tensor, Tensor]:
        return inputs, targets

    def _aggregate(self, result):
        return result

    def __call__(self, inp: Tensor, target: Tensor, **kwargs) -> Tensor:
        with torch.no_grad():
            return self._aggregate(self.forward(*self._prepare(*self._mask(inp, target)), **kwargs))


class MetricStack(Metric):
    """a metric stack is a keyed collection of metric functions that will are all to be called with\
    a given set of inputs.

    This is useful when you'd like to run more than 1 metric function on an output

    Note:
        if you're using :class:`hearth.loop.Loop` and pass a list of metrics a metric stack
        will be created for you automatically.

    Example:
        >>> import torch
        >>> from hearth.metrics import BinaryAccuracy, BinaryF1, MetricStack
        >>> _ = torch.manual_seed(0)
        >>>
        >>> metrics = MetricStack(BinaryAccuracy(), BinaryF1())
        >>> metrics
        MetricStack(BinaryAccuracy(mask_target=-1, from_logits=False),
                    BinaryF1(eps=1e-08, dim=0, mask_target=-1, from_logits=False))

        if metrics are provided as args keys will be created based on the metric names
        >>> list(metrics.keys())
        ['binary_accuracy', 'binary_f1']


        you can access the individual metric functions by key if you need to:

        >>> metrics.binary_accuracy
        BinaryAccuracy(mask_target=-1, from_logits=False)

        when calling the metric stack with inputs and targets will compute all metrics
        for those inputs and targets and return a keyed :class:`TensorDict`.

        >>> inputs = torch.rand(10, 1)
        >>> targets = torch.rand(10, 1).round()
        >>>
        >>> metrics(inputs, targets)
        TensorDict({'binary_accuracy': tensor(0.7000), 'binary_f1': tensor(0.5714)})


        if you would rather choose your own keys you can instantiate the MetricStack
        with keyword args like so:

        >>> metrics = MetricStack(acc=BinaryAccuracy(), f1=BinaryF1())
        >>> metrics(inputs, targets)
        TensorDict({'acc': tensor(0.7000), 'f1': tensor(0.5714)})


        keyword arguments will be passed to all metric functions
        this is useful if you need them for some metrics and not others
        particularly because all hearth metrics accept variable keyword args
        and ignore them.

        >>> def my_metric(inputs, targets, weights, **kwargs):
        ...     return (inputs * weights).sum() / (targets * weights).sum()
        >>>
        >>> metrics = MetricStack(BinaryAccuracy(), my_metric)
        >>> weights = torch.normal(0, 5, size=(10,1))
        >>> metrics(inputs, targets, weights=weights)
        TensorDict({'binary_accuracy': tensor(0.7000), 'my_metric': tensor(23.2894)})
    """

    def __init__(self, *args, **kwargs):
        self._fns = {}
        for arg in args:
            self._add_fn(arg)
        for k, v in kwargs.items():
            self._add_fn(v, k)

    def _add_fn(self, fn, key=None):
        if not callable(fn):
            raise TypeError(f'functions passed to {self.__class__.__name__} must be callable!')
        if key is None:
            key = fn.__name__ if hasattr(fn, '__name__') else fn.__class__.__name__
            key = to_snakecase(key)
        self._fns[key] = fn

    def items(self):
        return self._fns.items()

    def keys(self):
        return self._fns.keys()

    def __getattr__(self, k):
        try:
            return self._fns[k]
        except KeyError:
            raise AttributeError(k)

    def __len__(self):
        return len(self._fns)

    def __call__(self, inputs, targets, **kwargs):
        return TensorDict({k: func(inputs, targets, **kwargs) for k, func in self.items()})

    def __repr__(self):
        _argrepr = ', '.join(f'{f!r}' for f in self._fns.values())
        return f'{self.__class__.__name__}({_argrepr})'


class MultiHeadMetric(_MultiHeadFunc):
    """a wrapper for metrics multi-output models.

    Example:
        >>> import torch
        >>> from hearth.metrics import MultiHeadMetric, BinaryAccuracy, CategoricalAccuracy
        >>> _ = torch.manual_seed(0)
        >>>
        >>> metric = MultiHeadMetric(a=BinaryAccuracy(), b=CategoricalAccuracy())
        >>> metric
        MultiHeadMetric(a=BinaryAccuracy(mask_target=-1, from_logits=False),
                        b=CategoricalAccuracy(mask_target=-1))


        inputs and targets should be some kind of mapping with head names matching
        those we specified in our metric.

        output will be a :class:`hearth.containers.TensorDict`

        >>> batch_size = 10
        >>> inputs = {'a': torch.rand(batch_size, 1),
        ...           'b': torch.normal(batch_size, 1, size=(10, 4))}
        >>> targets = {'a': torch.rand(batch_size, 1).round(),
        ...            'b': torch.randint(4, size=(batch_size,))}
        >>> metric(inputs, targets)
        TensorDict({'a': tensor(0.4000), 'b': tensor(0.6000)})
    """

    def __call__(
        self, inputs: Mapping[str, Tensor], targets: Mapping[str, Tensor], **kwargs
    ) -> TensorDict:
        out = TensorDict()
        for k, func in self.items():
            out[k] = func(inputs[k], targets[k], **kwargs)
        return out



from typing import Tuple
import torch
from torch import Tensor


def accuracy(inputs: Tensor, targets: Tensor) -> Tensor:
    """basic accuracy based on equality.

    Args:
        inputs: binary values or class indices
        targets: ground truth binary values or class indices

    Returns:
        scalar tensor
    """
    return ((inputs == targets) * 1.0).mean()


def _onehot_inputs_and_targets(inputs: Tensor, targets: Tensor) -> Tuple[Tensor, Tensor]:
    expanded_targets = targets.unsqueeze(-1)
    fill = inputs.new_ones(expanded_targets.shape)
    pred = torch.zeros_like(inputs).scatter_add(-1, inputs.argmax(dim=-1, keepdim=True), fill)
    tru = torch.zeros_like(inputs).scatter_add(-1, expanded_targets, fill)
    return pred, tru


def precision(inputs: Tensor, targets: Tensor, eps=1e-8, dim=0):
    tp = inputs * targets
    return (tp.sum(dim=dim) + eps) / (inputs.sum(dim=dim) + eps)


def recall(inputs: Tensor, targets: Tensor, eps=1e-8, dim=0):
    tp = inputs * targets
    return (tp.sum(dim=dim) + eps) / (targets.sum(dim=dim) + eps)


def fbeta(inputs: Tensor, targets: Tensor, beta: float, dim=0, eps=1e-8):
    r = recall(inputs, targets, dim=dim, eps=eps)
    p = precision(inputs, targets, dim=dim, eps=eps)
    return ((1 + beta ** 2) * p * r) / (beta ** 2 * p + r)


def f1(inputs: Tensor, targets: Tensor, dim=0, eps=1e-8):
    return fbeta(inputs, targets, beta=1.0, dim=dim, eps=eps)



from typing import Union, List
import torch
from torch import nn
from hearth.modules import BaseModule


class LayerNormSimple(BaseModule):
    """Layer norm without learnable bias and gain.

    This basically just wraps the standard torch LayerNorm so it has no elementwise
    affine by default.

    Args:
        normalized_shape: input shape from an expected input. If a single integer is used, it is
             treated as a singleton list, and this module will normalize over the last dimension
             which is expected to be of that specific size.
        eps: a value added to the denominator for numerical stability. Default: 1e-5

    **Reference**:
        `Xu et al: Understanding and Improving Layer Normalization
        <https://arxiv.org/abs/1911.07013>`_

    Example:
        >>> import torch
        >>> from hearth.modules import LayerNormSimple
        >>>
        >>> layer = LayerNormSimple(5) # 5 feats
        >>> layer
        LayerNormSimple(5, eps=1e-05)

        >>> x = torch.tensor([[-4.2721, -2.8159, -1.2351,  0.2388,  4.5915],
        ...                  [-0.9092, -3.9666, -1.4216, -4.7373,  2.0403],
        ...                  [-1.2210,  4.4796, -1.2772, -2.8781,  4.1868]])
        >>> layer(x)
        tensor([[-1.1730, -0.6950, -0.1761,  0.3077,  1.7365],
                [ 0.3694, -0.9000,  0.1566, -1.2200,  1.5940],
                [-0.6139,  1.2486, -0.6323, -1.1554,  1.1530]])

    """

    def __init__(self, normalized_shape: Union[int, List[int], torch.Size], eps: float = 1e-5):
        super().__init__()
        self.norm = nn.LayerNorm(normalized_shape, eps=eps, elementwise_affine=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.norm(x)

    def __repr__(self) -> str:
        shape_or_feats = self.norm.normalized_shape
        if len(shape_or_feats) == 1:  # type: ignore
            shape_or_feats = shape_or_feats[0]  # type: ignore
        return f'{self.__class__.__name__}({shape_or_feats}, eps={self.norm.eps})'



import os
from dataclasses import dataclass
from typing import Union, Callable, Optional, Sequence, Tuple, List

import torch
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

import torchvision.transforms as T  # noqa: N812

from hearth.vision.utils import is_valid_filename, _image_file_end_expr, show_img, read_image
from hearth.vision.transforms import ResizeCrop


@dataclass
class RGBImageDataset(Dataset):
    """simple RGB image dataset that loads images from a directory. supports augmentations and \
    various utility methods as well as batch indexing (from list or slice).

    Images accessed with __getitem__ are read, resized, optionally augmented and finally transformed
    into float tensors in range ``[0,1]``.

    Args:
        directory: path to the directory to read images from. Homedir style paths are supported.
        size: integer or tuple size for image. If a ::attr::`size` is a single integer or
            a tuple of the same two values we will use
            :class:`hearth.vision.transforms.ResizeCrop` for resizing and
            :attr:`crop_noise` will be passed. If a tuple of different sizes are passed for
            non-square inputs :attr:`crop_noise` will be ignored and we will simply resize to the
            requested size. Defaults to 224 (commonly used in imagenet etc).
        crop_noise: noise: for shifting the crop window along the longest dimension.
            Defaults to 0.0 (no noise).
        augment: optional augmentation function which accepts a single image tensor of shape
            ``(C, W, H)`` of integers in range ``[0,255]`` returns augmented version of that image.
            Defaults to None (no augmentation).
        files: If a sequence of filenames is provided here we will use them instead of scanning the
            directory. The filenames are assumed to exist in the directory. This can be useful when
            breaking up a train and validation set which share the same directory or when working
            with a subset of the data. Defaults to None (directoy will be scanned.)

    """

    directory: str
    size: Union[int, Tuple[int, int]] = 224
    crop_noise: float = 0.0
    augment: Optional[Callable[[Tensor], Tensor]] = None
    files: Optional[Sequence[str]] = None

    def __post_init__(self):
        if self.files is None:
            self._files = list(
                filter(is_valid_filename, os.listdir(os.path.expanduser(self.directory)))
            )
        else:
            self._files = list(self.files)

        if isinstance(self.size, int):
            resize = ResizeCrop(self.size, self.crop_noise)
        elif self.size[0] == self.size[1]:
            resize = ResizeCrop(self.size[0], self.crop_noise)
        else:
            resize = T.Resize(self.size)
        self._resize = resize

    def _augment(self, img: Tensor) -> Tensor:
        if self.augment is None:
            return img
        return self.augment(img)

    def _build_path(self, filename: str) -> str:
        return os.path.join(self.directory, filename)

    def _get_filename(self, i: int) -> str:
        return self._files[i]

    def ids(self):
        """yields a list of ids (filenames stripped of endings)"""
        for f in self._files:
            yield _image_file_end_expr.sub('', f)

    def _load_img(self, filename, augment: bool = True):
        img = self._resize(read_image(self._build_path(filename)))
        if augment:
            img = self._augment(img)
        return img / 255

    def show_img(self, i: int, augment: bool = True):
        """display the image at index i with optional augmentation applied"""
        return show_img(self._load_img(self._files[i], augment=augment))

    def _get_batch(self, i: Union[slice, List[int]]):
        if isinstance(i, slice):
            filenames = self._files[i]
        else:
            filenames = [self._get_filename(ix) for ix in i]
        return torch.stack([self._load_img(f) for f in filenames])

    def _get_image(self, i: int):
        return self._load_img(self._files[i])

    def __getitem__(self, i: Union[int, slice, List[int]]):
        if isinstance(i, int):
            return self._get_image(i)
        return self._get_batch(i)

    def __len__(self) -> int:
        return len(self._files)

    def batches(
        self,
        batch_size: int = 32,
        shuffle: bool = False,
        drop_last: bool = False,
        num_workers: int = 0,
        **kwargs,
    ) -> DataLoader:
        """return a Dataloader that iterates over batches of this dataset.

        Note:
            this method supports additional keyword args which will be passed to the dataloader.

        Args:
            batch_size: Defaults to 32.
            shuffle: if True shuffle the dataset. Defaults to False.
            drop_last: if true drop the last batch if it is less than batch size. Defaults to False.
            num_workers: number of workers. Defaults to 0.
        """
        return DataLoader(
            self,
            batch_size=batch_size,
            num_workers=num_workers,
            drop_last=drop_last,
            shuffle=shuffle,
            **kwargs,
        )



import sys

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal
from dataclasses import dataclass
import operator
from functools import reduce, partial
from hearth.callbacks.base import Callback
from hearth.events import Improvement, Stagnation


def dotted_attrgetter(path: str, obj):
    return reduce(getattr, path.split('.'), obj)


@dataclass
class ImprovementMonitor(Callback):
    """This callback monitors a metric or loss on the specified stage for improvement and \
    stagnation.

    When improvement or stagnation is detected on the metric it emits \
     :class:`hearth.events.Improvement` and :class:`hearth.events.Stagnation` events which can be \
    used in other callbacks.

    Args:
        field: the field to monitor on the loop... may be a dotted path string for nested objects.
            Defaults to 'loss'.
        improvement_on: string operator specifier ie (lt is less_than and gt is greater than).
            Defaults to 'lt' (for loss measurement).
        stage: named stage to measure improvement on. this should correspond to stages on your
            loop obviously. Defaults to 'val'.
        stagnant_after: wait this number of steps before you start issuing stagnation events.
            Defaults to 1.

    **Active On:**
        - stage_end

    **Events Emitted:**
        - :class:`hearth.events.Improvement`
        - :class:`hearth.events.Stagnation`

    **Accesses Loop Attributes:**
        - metric or loss
        - stage
        - epoch
    """

    field: str = 'loss'
    improvement_on: Literal['gt', 'lt'] = 'lt'
    stage: str = 'val'
    stagnant_after: int = 1

    def __post_init__(self):
        if self.improvement_on not in ('gt', 'lt'):
            raise ValueError(
                f'improvement_on must be one of ["gt", "lt"] but got {self.improvement_on}'
            )
        self._get_value = partial(dotted_attrgetter, self.field)
        self._is_improvement = getattr(operator, self.improvement_on)
        self._last_best = float('inf') if self.improvement_on == 'lt' else -float('inf')
        self._best_step = -1

    def on_stage_end(self, loop):
        if loop.stage == self.stage:
            this_value = self._get_value(loop)
            steps = loop.epoch - self._best_step
            event = None
            if self._is_improvement(this_value, self._last_best):
                event = Improvement(
                    self.field,
                    stage=self.stage,
                    steps=steps,
                    best=this_value,
                    last_best=self._last_best,
                )
                self._best_step = loop.epoch
                self._last_best = this_value
            elif steps > self.stagnant_after:
                event = Stagnation(
                    field=self.field, stage=self.stage, steps=steps, best=self._last_best
                )
            if event:
                loop.fire(event)



from typing import Sequence, Type, Optional, Callable
import os
from dataclasses import dataclass
from copy import deepcopy
import torch
from hearth.events import MonitoringEvent
from hearth.callbacks import Callback
from hearth.events import Improvement, CheckpointSaved
from hearth.modules import BaseModule
from hearth._file_utils import mkdirs_if_not_exist


@dataclass
class Checkpoint(Callback):
    """This callback saves checkpoints on certain events.

    Note:
        only models derived from :class:`hearth.modules.BaseModule` are currently supported.

    Args:
        model_dir: directory to save model checkpoint in. If the directory does not exist
            it will be created on registration.
        event_types: Types of monitoring events to checkpoint on.
            Default is (:class:`hearth.events.Improvement`, )
        prepare_model: Optional callable which accepts and returns
             a :class:`BaseModule` for preparing the model to be saved.
             This function will always recieve a **copy** of the model on the loop just for
             safety. Defaults to None.
        field: If provided only save on events where field matches this field.
        stage: If provided only save on events where stage matches this stage.
        save_history: if True save the loop history at this step to the model dir. Defaults to True
        save_optimizer: if True save the optimizer state dict to the model dir. Defaults to True.

    **Active On:**
        - registration
        - event
        - epoch_end

    **Events Listened For:**
        - :class:`hearth.events.Improvement` (default)

    **Events Emitted:**
        - :class:`hearth.events.ModelSaved`

    **Accesses Loop Attributes:**
        - model
        - history (if save_history is True)
        - optimizer (if save_optimizer is True)

    **Accesses Event Attributes:**
        - field
        - stage
    """

    model_dir: str
    event_types: Sequence[Type[MonitoringEvent]] = (Improvement,)
    prepare_model: Optional[Callable[[BaseModule], BaseModule]] = None
    field: Optional[str] = None
    stage: Optional[str] = None
    save_history: bool = True
    save_optimizer: bool = True

    def __post_init__(self):
        self._should_save = False

    def on_registration(self, loop):
        # check model
        if not isinstance(loop.model, BaseModule):
            raise TypeError(f'{self.__class__.__name__} callback only supports hearth.BaseModule')
        if os.path.exists(self.model_dir):
            if not os.path.isdir(self.model_dir):
                raise NotADirectoryError(
                    f'{self.__class__.__name__} expects model_dir to be a directory!'
                )
        else:
            mkdirs_if_not_exist(self.model_dir, verbose=True)

    def _is_save_event(self, event):
        if isinstance(event, self.event_types):
            if self.field:
                if event.field != self.field:
                    return False
            if self.stage:
                if event.stage != self.stage:
                    return False
            return True
        return False

    def _save_model(self, model):
        save_model = deepcopy(model)
        if self.prepare_model is not None:
            save_model = self.prepare_model(save_model)
        save_model.save(self.model_dir)

    def _save_history(self, history):
        history.save(self.model_dir)

    def _save_optimizer(self, optimizer):
        path = os.path.join(self.model_dir, 'optimizer_state.pt')
        torch.save(optimizer.state_dict(), path)

    def save_checkpoint(self, loop):
        self._save_model(loop.model)
        if self.save_history:
            self._save_history(loop.history)
        if self.save_optimizer:
            self._save_optimizer(loop.optimizer)

    def on_epoch_end(self, loop):
        if self._should_save:
            self.save_checkpoint(loop)
            event = CheckpointSaved(self.model_dir)
            loop.fire(event)
            self._should_save = False

    def on_event(self, loop, event):
        if self._is_save_event(event):
            self._should_save = True



from typing import Union


class Callback:
    """Base class for all callbacks.

    Child callbacks should subclass this and override methods they want to actually do something on.
    all methods will be passed the loop  and the ``on_event`` method will additonally be passed an
    event.
    """

    active: bool = True

    def on_registration(self, loop):
        """this will be called when the loop sets up the callbacks before any training starts."""
        pass

    def on_stage_start(self, loop):
        """This will be called on the start of a stage.

        an identifier such as `train`, `val` etc will be accessable at ``loop.stage``.
        """
        pass

    def on_stage_end(self, loop):
        """This will be called on the start of a stage.

        an identifier such as `train`, `val` etc will be accessable at ``loop.stage``.
        """
        pass

    def on_epoch_start(self, loop):
        """This will be called on when each epoch starts."""
        pass

    def on_epoch_end(self, loop):
        """This will be called on when each epoch ends."""
        pass

    def on_batch_start(self, loop):
        """this will be called before the batch is passed to the model on each stage"""
        pass

    def on_batch_end(self, loop):
        """this will be called after each batch is proccessed (for each stage)"""
        pass

    def on_loss_start(self, loop):
        """this will be called for each batch before loss is calculated (for each stage)"""
        pass

    def on_loss_end(self, loop):
        """this will be called for each batch after the loss is calculated (for each stage)"""
        pass

    def on_step_start(self, loop):
        """this will be called for before the optimizer step (for training stage only)"""
        pass

    def on_step_end(self, loop):
        """this will be called after the optimizer step (for training stage only)"""
        pass

    def on_metric_start(self, loop):
        """this will be called just before metrics are calculated (for each batch)"""
        pass

    def on_metric_end(self, loop):
        """this will be called just after are metrics calculated (for each batch)"""
        pass

    def on_backward_start(self, loop):
        """this will be called just before backward is called on the loss (for each batch)"""
        pass

    def on_backward_end(self, loop):
        """this will be called just after backward is called (for each batch)"""
        pass

    def on_event(self, loop, event):
        """this will be called whenever an event is fired"""
        pass

    def _repr_args(self) -> str:
        return ''

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._repr_args()})'


class CallbackManager(Callback):
    """Manages multiple callbacks and calls them in order."""

    def __init__(self, *callbacks: Callback):
        self._callbacks = []
        for callback in callbacks:
            if not isinstance(callback, Callback):
                raise TypeError(f'expected Callback type but got {type(callback)}.')

            self._callbacks.append(callback)

    def on_registration(self, loop):
        for callback in self:
            callback.on_registration(loop)

    def on_stage_start(self, loop):
        for callback in self:
            callback.on_stage_start(loop)

    def on_stage_end(self, loop):
        for callback in self:
            callback.on_stage_end(loop)

    def on_epoch_start(self, loop):
        for callback in self:
            callback.on_epoch_start(loop)

    def on_epoch_end(self, loop):
        for callback in self:
            callback.on_epoch_end(loop)

    def on_batch_start(self, loop):
        for callback in self:
            callback.on_batch_start(loop)

    def on_batch_end(self, loop):
        for callback in self:
            callback.on_batch_end(loop)

    def on_loss_start(self, loop):
        for callback in self:
            callback.on_loss_start(loop)

    def on_loss_end(self, loop):
        for callback in self:
            callback.on_loss_end(loop)

    def on_step_start(self, loop):
        for callback in self:
            callback.on_step_start(loop)

    def on_step_end(self, loop):
        for callback in self:
            callback.on_step_end(loop)

    def on_metric_start(self, loop):
        for callback in self:
            callback.on_metric_start(loop)

    def on_metric_end(self, loop):
        for callback in self:
            callback.on_metric_end(loop)

    def on_backward_start(self, loop):
        for callback in self:
            callback.on_backward_start(loop)

    def on_backward_end(self, loop):
        for callback in self:
            callback.on_backward_end(loop)

    def on_event(self, loop, event):
        for callback in self:
            callback.on_event(loop, event)

    def _repr_args(self) -> str:
        return ','.join(f'{callback!r}' for callback in self._callbacks)

    def __len__(self) -> int:
        return len(self._callbacks)

    def __getitem__(self, i: Union[int, slice]) -> Callback:
        if isinstance(i, slice):
            return CallbackManager(*self._callbacks[i])
        return self._callbacks[i]  # type: ignore



from typing import Type
import torch
from torch import nn
from hearth._internals import Registry

_activation_registry: Registry[Type[nn.Module]] = Registry('activations')
_activation_registry.register(nn.ReLU)
_activation_registry.register(nn.RReLU)
_activation_registry.register(nn.Hardtanh)
_activation_registry.register(nn.ReLU6)
_activation_registry.register(nn.Sigmoid)
_activation_registry.register(nn.Hardsigmoid)
_activation_registry.register(nn.Tanh)
_activation_registry.register(nn.SiLU)
_activation_registry.register(nn.Hardswish)
_activation_registry.register(nn.ELU)
_activation_registry.register(nn.CELU)
_activation_registry.register(nn.SELU)
_activation_registry.register(nn.GLU)
_activation_registry.register(nn.GELU)
_activation_registry.register(nn.Hardshrink)
_activation_registry.register(nn.LeakyReLU)
_activation_registry.register(nn.LogSigmoid)
_activation_registry.register(nn.Softplus)
_activation_registry.register(nn.Softshrink)
_activation_registry.register(nn.PReLU)
_activation_registry.register(nn.Softsign)
_activation_registry.register(nn.Tanhshrink)


def mish(x: torch.Tensor) -> torch.Tensor:
    """functional version of mish activation see :class:`Mish` for more info."""
    return x * torch.tanh(torch.nn.functional.softplus(x))


@_activation_registry.register
class Mish(nn.Module):
    """Applies the Mish activation function element-wise

    :math:`\\text{Mish}(x)=x\\tanh(softplus(x))`

    .. image:: ../images/mish.png

    **Reference**:
        `Diganta Misra: Mish: A Self Regularized Non-Monotonic Activation Function
        <https://arxiv.org/abs/1908.08681>`_

    Example:
        >>> import torch
        >>> from hearth.activations import Mish
        >>>
        >>> activation = Mish()
        >>> x = torch.linspace(-2, 2, 8)
        >>> activation(x)
        tensor([-0.2525, -0.3023, -0.2912, -0.1452,  0.1969,  0.7174,  1.3256,  1.9440])
    """

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return mish(x)


def get_activation(name: str, **kwargs) -> nn.Module:
    """get an new instance activiation by its name, optionally passing any extra kwargs.

    Args:
        name: string name of the activation. will be best effort normalized.

    Returns:
        nn.Module: instance of an activation module

    Example:
        >>> from torch import nn
        >>> from hearth.activations import get_activation
        >>>
        >>> get_activation('relu')
        ReLU()

        you can also pass kwargs... although this is slightly silly
        since you may as well use the actual module then...

        >>> get_activation('prelu', num_parameters=6)
        PReLU(num_parameters=6)

        name will be normalized so dont worry about casing:

        >>> get_activation('mish')
        Mish()

        >>> get_activation('MISH')
        Mish()
    """
    return _activation_registry[name](**kwargs)  # type: ignore



import torch
from torch import nn
from hearth.modules import BaseModule


class Residual(BaseModule):
    """wraps a block in a residual connection :math:`y = block(x) + x`.

    Args:
        block: the module to wrap.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> from hearth.modules import Residual
        >>> _ = torch.manual_seed(0)
        >>>
        >>> res = Residual(nn.Linear(4, 4))
        >>>
        >>> x = torch.rand(2, 4) # (batch, feats)
        >>> res(x)
        tensor([[ 0.6371,  1.5493,  0.0031, -0.0379],
                [ 0.3584,  0.8512,  0.5208, -0.7607]], grad_fn=<AddBackward0>)
    """

    def __init__(self, block: nn.Module):
        super().__init__()
        self.block = block

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """forward padd for ``Residual`` wrapper."""
        y = self.block(x)
        return x + y


class ReZero(Residual):
    """Implements ReZero residual connection around a block with dropout (as in transformer\
     implementation).

    **Reference**:
        `Bachlechner et al: ReZero is All You Need: Fast Convergence at Large Depth
        <https://arxiv.org/abs/2003.04887>`_

    Example:
        >>> import torch
        >>> from hearth.modules import ReZero
        >>>
        >>> transformation = nn.Sequential(nn.Linear(10, 10),
        ...                                nn.ReLU(),
        ...                                nn.Dropout(.1),
        ...                                nn.Linear(10, 10)
        ...                                )
        >>> re_zero = ReZero(transformation, dropout=.1)
        >>>
        >>> x = torch.normal(0, 1, size=(5, 10))
        >>> y = re_zero(x)
        >>> y.shape
        torch.Size([5, 10])

        since ``re_zero``'s weight parameter has not actually been trained ``y`` its going to be
        equal to ``x`` and nothing from the transformation will be added to the input...
        as training goes on this should change.

        >>> (y == x).all()
        tensor(True)
    """

    def __init__(self, block: nn.Module, dropout: float = 0.0):
        super().__init__(block=block)
        self.dropout = nn.Dropout(dropout)
        self.res_weight = nn.Parameter(torch.tensor([0.0]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """forward for ReZero."""
        y = self.block(x)
        y = y * self.res_weight
        return x + self.dropout(y)



from .base import Callback, CallbackManager
from .logging import PrintLogger
from .grad_clipping import ClipGradNorm, ClipGradValue
from .monitors import ImprovementMonitor
from .checkpoints import Checkpoint
from .stopping import EarlyStopping
from .history import History
from .finetuner import FineTuneCallback
from .learning_rate import (
    LambdaLRCallback,
    MultiStepLRCallback,
    MultiplicativeLRCallback,
    StepLRCallback,
    ExponentialLRCallback,
    CosineAnnealingLRCallback,
    ReduceLROnPlateauCallback,
)

__all__ = [
    'Callback',
    'CallbackManager',
    'PrintLogger',
    'ClipGradNorm',
    'ClipGradValue',
    'ImprovementMonitor',
    'Checkpoint',
    'EarlyStopping',
    'FineTuneCallback',
    'History',
    'LambdaLRCallback',
    'MultiStepLRCallback',
    'MultiplicativeLRCallback',
    'StepLRCallback',
    'ExponentialLRCallback',
    'CosineAnnealingLRCallback',
    'ReduceLROnPlateauCallback',
]



from dataclasses import dataclass


@dataclass
class Event:
    """base class for all events."""

    @property
    def msg(self) -> str:
        return ''

    def logmsg(self) -> str:
        return f'{self.__class__.__name__}{self.msg}'


@dataclass
class MonitoringEvent(Event):
    """base class for monitoring events"""

    field: str
    stage: str
    steps: int

    def _get_stepmsg(self) -> str:
        return f'{self.steps} step{"s" if self.steps > 1 else ""}'

    def _get_fieldmsg(self) -> str:
        return f'{self.stage}.{self.field}'

    def logmsg(self) -> str:
        return f'{self.__class__.__name__}[{self._get_fieldmsg()}] {self.msg}'


@dataclass
class Improvement(MonitoringEvent):
    """This event should be emitted on an monitored improvement.

    Args:
        field: the field being monitored.
        stage: the stage being monitored.
        steps: the number of steps (generally epochs) that between this and the last improvement.
        best: the best value.
        last_best: the last best value
    """

    best: float
    last_best: float

    @property
    def msg(self) -> str:
        return (
            f'improved from : {self.last_best:0.4f} to {self.best:0.4f} in {self._get_stepmsg()}.'
        )


@dataclass
class Stagnation(MonitoringEvent):
    """This event should be emitted when a monitored metric or loss is not improving.

    Args:
        field: the field being monitored.
        stage: the stage being monitored.
        steps: the number of steps (generally epochs) that the measured value has been stagnant.
        best: the best value seen so far.
    """

    best: float

    @property
    def msg(self) -> str:
        return f'stagnant for {self._get_stepmsg()}' f' no improvement from {self.best:0.4f}.'


@dataclass
class CheckpointSaved(Event):
    """This event should be emitted when a model is saved as part of checkpointing in a callback.

    Args:
        model_dir: the directory the model was saved to.
    """

    model_dir: str

    @property
    def msg(self):
        return f' checkpoint saved to {self.model_dir}.'


@dataclass
class EarlyStop(Event):
    """this event should be emitted when training is instructed to stop early.

    Args:
        epoch: the epoch that stopping is triggered.
    """

    epoch: int

    @property
    def msg(self):
        return f' triggered at epoch {self.epoch}.'


@dataclass
class UnbottleEvent(Event):
    """this event should be emitted when fintuning.

    Args:
        epoch: the epoch that stopping is triggered.
    """

    epoch: int
    block: str
    lr: float

    @property
    def msg(self):
        return f' Unbottled block {self.block} with learning rate {self.lr} on epoch {self.epoch}.'


@dataclass
class UnbottlingComplete(Event):
    """this event should be emitted when unbottling is complete during finetuning.

    Args:
        epoch: the epoch that stopping is triggered.
    """

    @property
    def msg(self):
        return ' Unbottling complete nothing left to unbottle.'



import os
from typing import Dict
from pprint import pformat
from collectionish import AttyDict
from collections import deque
from hearth.callbacks import Callback
from hearth._file_utils import load_json, save_json


class History(Callback):
    """This callback tracks history across epochs.

    Note:
        this callback is auto-registered :class:`hearth.loop.Loop` by default so you
        don't need to list it in callbacks. It should be accessable at ``loop.history``

    **Active On:**
        - epoch_start
        - stage_end
        - epoch_end

    **Accesses Loop Attributes:**
        - loss
        - metrics
        - epoch
        - optimizer
        - stage
    """

    @classmethod
    def load(cls, model_dir) -> 'History':
        """load this history object from the model_dir"""
        path = os.path.join(model_dir, 'history.json')
        hist = load_json(path)
        return cls(*hist)

    def __init__(self, *history):
        self._history = [AttyDict(row) for row in history]
        self._current_step_buffer = deque(maxlen=1)

    @property
    def current_step(self):
        """the current step if tracking is in the middle of an epoch."""
        return self._current_step_buffer[0]

    @property
    def last_epoch(self) -> int:
        """the last epoch in the complete history"""
        if len(self):
            return self[-1].epoch
        return -1

    def _gather_lrs(self, optimizer) -> Dict[str, float]:
        return {f'group{i}': group['lr'] for i, group in enumerate(optimizer.param_groups)}

    def on_epoch_start(self, loop):
        current_step = AttyDict(epoch=loop.epoch, lrs=self._gather_lrs(loop.optimizer))
        self._current_step_buffer.append(current_step)

    def on_stage_end(self, loop):
        self.current_step[loop.stage] = AttyDict(loss=loop.loss, metric=loop.metric)

    def on_epoch_end(self, loop):
        self._history.append(self._current_step_buffer.pop())

    def __len__(self) -> int:
        return len(self._history)

    def __getitem__(self, i):
        return self._history[i]

    def save(self, model_dir: str):
        """save this history to a file history.json in ``model_dir```."""
        path = os.path.join(model_dir, 'history.json')
        save_json(self._history, path)

    def _iter_pformat_args(self):
        if not self._history:
            yield ''
        else:
            padding = len(self.__class__.__name__)
            lines = pformat(self._history).split('\n')
            lines[-1] = lines[-1][:-1]
            yield lines.pop(0)[1:]
            for line in lines:
                yield ' ' * padding + line

    def __repr__(self):
        reprargs = '\n'.join(self._iter_pformat_args())
        return f'{self.__class__.__name__}({reprargs})'

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._history == other._history
        return NotImplemented



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()


test_requirements = ['pytest']

setup(
    author="Lea Provenzano",
    author_email='leaprovenzano@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3 :: Only',
    ],
    description="a collection tools and extensions for working with pytorch",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='hearth',
    name='hearth',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/leaprovenzano/hearth',
    version='0.2.0',
    zip_safe=False,
)



import os
from typing import Iterator
import torch
from torch import nn
from hearth.grad import freeze, unfreeze, trainable_parameters, allgrad
from hearth._config import _init_wrapper, from_config
from hearth._file_utils import save_json, load_json


class BaseModule(nn.Module):
    """A base class like nn.Module but with a few extra useful bits.

    **features include**:
        - auto config generation based on init args.
        - standardized loading and saving.
        - torchscripting via the :meth:`script` method.
        - whole model freezing via the :meth:`freeze` method.
        - and more...
    """

    @classmethod
    def from_config(cls, config):
        """given a valid config return a new instance of this Module."""
        return from_config(cls, config)

    @classmethod
    def load(cls, model_dir: str, strict: bool = True):
        """create a new instance of this with config and parameters loaded from  ``model_dir``.

        This method expects model dir to have the following files:
             - state.pt: the state dict for this model
             - config.json :  the config of this model, nessisary to reinstantiate a new model \
                 with the :meth:`from_config`  method.

         you can use the :meth:`save` method on an instance of this module to create them.

         Args:
             model_dir: directory to load from.
             strict: if we shold be strict about loading the state dict. Defaults to True.
        """
        config = load_json(os.path.join(model_dir, 'config.json'))
        state_dict = torch.load(os.path.join(model_dir, 'state.pt'))

        instance = cls.from_config(config)
        instance.load_state_dict(state_dict, strict=strict)
        return instance

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        cls.__init__ = _init_wrapper(cls.__init__)

    def freeze(self):
        """freeze this model in place."""
        freeze(self)

    def unfreeze(self):
        """unfreeze this model in place."""
        unfreeze(self)

    def script(self) -> torch.jit.RecursiveScriptModule:
        """torchscript this model using jit.script."""
        with torch.jit.optimized_execution(True):
            scripted = torch.jit.script(self)
        return scripted

    def trainable_parameters(self) -> Iterator[nn.Parameter]:
        """yields trainable parameters from this model."""
        yield from trainable_parameters(self)

    def config(self):
        """get the config for this module.

        This config can be passed to the :meth:`from_config` class method to create a new
        instance of this module.
        """
        return dict(self.__config__)

    def save(self, model_dir: str):
        """save this models state and config to a ``model_dir`` so it can be re-created later.

        This method will generate the following files in the ``model_dir``:
            - state.pt: the state dict for this model
            - config.json :  the config of this model, nessisary to reinstantiate a new model \
                with the :meth:`from_config`  method.

        Args:
            model_dir: directory to save stuff in.
        """
        torch.save(self.state_dict(), os.path.join(model_dir, 'state.pt'))
        save_json(self.config(), path=os.path.join(model_dir, 'config.json'))

    def blocks(self) -> Iterator[nn.Module]:
        """this override this method to define depth based sections of your network.

        Defining :meth:`blocks` is useful when your'e doing finetuning or want to use depth
        based learning rates. It's how methods like :meth:`bottleneck` and :meth:`unbottle` know
        how to iterate over your model. When overriden this method should yield logical depth based
        sections of your model, which should themselves be :class:`nn.Module`s **in depth based
        order from input to output**. How you section your network or group things together is up to
        you. If not overridden this method will just yield `self`
        """
        yield self

    def reverse_blocks(self) -> Iterator[nn.Module]:
        """yields blocks from this modules :meth:`blocks` method in reverse order (from output to\
         input.)
        """
        yield from list(self.blocks())[::-1]

    def depth(self) -> int:
        """the block depth of this module (as defined in :meth:`blocks`)."""
        return len(list(self.blocks()))

    def bottleneck(self, n: int):
        depth = self.depth()
        if n > depth:
            raise ValueError(f'cannot bottleneck {n} blocks where depth is only {depth}')
        for i, block in enumerate(self.blocks()):
            if i < n:
                freeze(block)

    def unbottle(self):
        """unfreeze (inplace) a the next deepest block that is wholly or partially frozen.

        Returns:
            Union[None, nn.Module]: returns the unfrozen block if there was anything to unbottle
                otherwise None
        """
        for block in self.reverse_blocks():
            if not allgrad(block):
                return unfreeze(block)
        return None



from dataclasses import dataclass
from hearth.callbacks import Callback
from hearth.events import Event


DEFAULT_BATCH_FMT = (
    'epoch: {loop.epoch} stage: [{loop.stage}]'
    ' batch: {loop.batches_seen}/{loop.n_batches}'
    ' loss: {loop.loss:0.4f}'
)

DEFAULT_METRIC_FMT = " {loop.metric:0.4f}"


@dataclass
class PrintLogger(Callback):
    """a very simple logging callback that just prints stuff to sdout.

    Args:
        batch_format: format string which will printed (single line) for each batch
            and will passed a single argument ``loop``. Defaults to \
            `hearth.callbacks.logging.DEFAULT_BATCH_FMT`.
        epoch_delim: single char delimiter that will be used to seperate epochs. Defaults to ``-``.
        epoch_delim_width: width of epoch delimiter. Defaults to 80.

    Example:
        >>> import torch
        >>> from torch import nn
        >>> _ = torch.manual_seed(0)
        >>> from torch.utils.data import TensorDataset, DataLoader
        >>> from hearth.loop import Loop
        >>> from hearth.callbacks import PrintLogger
        >>> from hearth.metrics import BinaryAccuracy
        >>>
        >>> train = TensorDataset(torch.normal(0, 2, size=(5000, 64)),
        ...                       torch.randint(2, size=(5000, 1))*1.0)
        >>> val = TensorDataset(torch.normal(0, 2, size=(3000, 64)),
        ...                     torch.randint(2, size=(3000, 1))*1.0)
        >>>
        >>> train_batches = DataLoader(train, batch_size=32, shuffle=True, drop_last=False)
        >>> val_batches = DataLoader(val, batch_size=32, shuffle=True, drop_last=False)
        >>>
        >>> model = nn.Sequential(nn.Linear(64, 128), nn.ReLU(), nn.Linear(128, 1), nn.Sigmoid())
        >>>
        >>> loop = Loop(model=model,
        ...         optimizer=torch.optim.AdamW(model.parameters(), lr=0.001),
        ...        loss_fn = nn.BCELoss(),
        ...        metrics = BinaryAccuracy(),
        ...        callbacks= [PrintLogger()]
        ...       )
        >>> loop(train_batches, val_batches, 2) # doctest: +SKIP
        epoch: 0 stage: [train] batch: 157/157 loss: 0.7036 metric: 0.5054
        epoch: 0 stage: [val] batch: 94/94 loss: 0.7036 metric: 0.4933
        --------------------------------------------------------------------------------
        epoch: 1 stage: [train] batch: 157/157 loss: 0.6799 metric: 0.5660
        epoch: 1 stage: [val] batch: 94/94 loss: 0.7097 metric: 0.4860
        --------------------------------------------------------------------------------
    """

    batch_format: str = DEFAULT_BATCH_FMT
    metric_format: str = DEFAULT_METRIC_FMT
    epoch_delim: str = '-'
    epoch_delim_width: int = 80

    def __post_init__(self):
        self._end_cursor = None

    def print_msg(self, msg: str):
        print(msg, end=self._end_cursor, flush=True, sep='')

    def on_registration(self, loop):
        if loop._has_metrics:
            self.batch_format = self.batch_format + self.metric_format

    def get_batch_msg(self, loop) -> str:
        return self.batch_format.format(loop=loop)

    def on_epoch_end(self, loop):
        print(self.epoch_delim * self.epoch_delim_width)

    def on_batch_start(self, loop):
        if loop.n_batches and (loop.batches_seen + 1 == loop.n_batches):
            self._end_cursor = None
        else:
            self._end_cursor = '\r'

    def on_batch_end(self, loop):
        return self.print_msg(self.get_batch_msg(loop))

    def on_stage_end(self, loop):
        self._end_cursor = None

    def on_event(self, loop, event: Event):
        print(f'EVENT: {event.logmsg()}')
